use axum::{
    extract::{Path, State},
    http::{header, StatusCode},
    response::{Html, IntoResponse},
    routing::get,
    Router,
};
use std::{collections::HashMap, path::PathBuf, sync::Arc};
use clap::Parser;

#[derive(Debug)]
struct BlogState {
    templates: HashMap<String, Cached<Template>>,
    posts: HashMap<String, Cached<Post>>,

    config: Config,
}
type AppState = State<Arc<BlogState>>;

#[derive(Debug)]
struct Cached<S> {
    data: S,
    path: PathBuf,
    // TODO
    // last_read: NaiveDateTime,
}

impl<S: From<String>> Cached<S> {
    fn data(&self) -> std::io::Result<S> {
        let contents = std::fs::read_to_string(&self.path)?;
        Ok(S::from(contents))

        // TODO
        //if self.last_read + Duration::minutes(30) > now {
        //    let contents = std::fs::read_to_string(&self.path)?;
        //
        //    self.last_read = now;
        //    self.data = S::from(contents);
        //
        //    Ok(&self.data)
        //} else {
        //    Ok(&self.data)
        //}
    }

    fn from_file_path(path: PathBuf) -> std::io::Result<Self> {
        let contents = std::fs::read_to_string(&path)?;

        Ok(Self {
            data: S::from(contents),
            path,
            // last_read: Utc::now().naive_utc(),
        })
    }
}

#[derive(Debug)]
struct Template {
    index: usize,
    content: String,
}

impl From<String> for Template {
    fn from(value: String) -> Self {
        let index = value.find("%%%").unwrap_or_default();

        Self {
            index,
            content: value,
        }
    }
}

#[derive(Debug)]
struct Post {
    content: String,
    embed_template_name: Option<String>,
}

impl Post {
    fn to_html(&self, templates: &HashMap<String, Cached<Template>>) -> String {
        let mut s = String::with_capacity(self.content.len());
        if let Some(template) = self
            .embed_template_name
            .as_ref()
            .and_then(|name| templates.get(name))
        {
            if template.data.index > 0 {
                s += &template.data.content[0..template.data.index];
            }

            s += &markdown::to_html(&self.content);

            if template.data.index > 0 {
                s += &template.data.content[template.data.index + 3..]
            }
        } else {
            s += &markdown::to_html(&self.content);
        }

        s
    }
}

impl From<String> for Post {
    fn from(value: String) -> Self {
        let template_name = if value.starts_with("%{") {
            value[2..].split('}').next().map(Into::into)
        } else {
            None
        };

        Self {
            content: if template_name.is_some() {
                value.lines().skip(1).collect::<Vec<_>>().join("\n")
            } else {
                value
            },
            embed_template_name: template_name,
        }
    }
}

#[derive(clap::Parser, Debug)]
struct Config {
    #[clap(env, short, long)]
    port: u16,
    #[clap(env, short, long)]
    blog_dir: String,
}

#[tokio::main]
async fn main() {
    let config = Config::parse();
    let mut blog_state = BlogState {
        templates: HashMap::new(),
        posts: HashMap::new(),
        config,
    };

    for dir in std::fs::read_dir(&blog_state.config.blog_dir).unwrap().flatten() {
        if dir.file_type().unwrap().is_file() {
            let file_name = dir.file_name().into_string().unwrap();
            let Some((file_name, extension)) = file_name.split_once('.') else {
                continue;
            };

            match extension {
                "html.template" => {
                    blog_state.templates.insert(
                        file_name.to_string(),
                        Cached::from_file_path(dir.path()).unwrap(),
                    );
                }
                "md" => {
                    blog_state.posts.insert(
                        file_name.to_string(),
                        Cached::from_file_path(dir.path()).unwrap(),
                    );
                }
                _ => {}
            }
        }
    }

    let port = blog_state.config.port;
    let app = Router::new()
        .route("/", get(home_page))
        .route("/:post", get(blog_post))
        .route("/images/:image", get(image))
        .with_state(Arc::new(blog_state));
    let listener = tokio::net::TcpListener::bind(format!("0.0.0.0:{}", port)).await.unwrap();

    axum::serve(listener, app).await.unwrap();
}

async fn home_page(State(state): AppState) -> Result<Html<String>, Html<String>> {
    match state.posts.get("home") {
        Some(post) => match post.data() {
            Ok(content) => Ok(Html(content.to_html(&state.templates))),
            Err(e) => Err(Html(e.to_string())),
        },
        None => Err(Html("<h1>404 - Not Found</h1>".to_string())),
    }
}

async fn blog_post(
    State(state): AppState,
    Path(post): Path<String>,
) -> Result<Html<String>, Html<String>> {
    match state.posts.get(&post) {
        Some(post) => match post.data() {
            Ok(content) => Ok(Html(content.to_html(&state.templates))),
            Err(e) => Err(Html(e.to_string())),
        },
        None => Err(not_found_page(state)),
    }
}

async fn image(State(state): AppState, Path(image_path): Path<String>) -> impl IntoResponse {
    match std::fs::read(format!("{}/images/{image_path}", state.config.blog_dir)) {
        Ok(contents) => Ok(([(header::CONTENT_TYPE, "image/png")], contents)),
        Err(e) => {
            eprintln!("{e:#?}");
            Err(StatusCode::NOT_FOUND)
        }
    }
}

fn not_found_page(state: Arc<BlogState>) -> Html<String> {
    match state.posts.get("404") {
        Some(post) => match post.data() {
            Ok(content) => Html(content.to_html(&state.templates)),
            Err(e) => Html(e.to_string()),
        },
        None => Html("<h1>404 - Not Found</h1>".to_string()),
    }
}
